#include <string>
#include <vector>

class Utf8ToLatin1 {
    
public:

    Utf8ToLatin1(char const *input, int recover)
    :   ptr((unsigned char const *) input),
        rec(recover),
        valid(true) {
    }

    void convert(char const *&begin, char const *&end) {
        while (convert_seq()) {
            // empty
        }
        static char const *emptyStr = "";
        if (valid) {
            if (out.empty()) {
                begin = emptyStr;
                end = emptyStr;
            } else {
                begin = (char const *) &out[0];
                end = begin + out.size();
            }
        } else {
            begin = NULL;
            end = NULL;
        }
    }

    char const *error() const {
        if (invalidUnicode >= 0) {
            static char msg[128];
            sprintf(msg, "utf8tolatin1: character not latin-1 (U+%X)", invalidUnicode);
            return msg;
        } else {
            return "utf8tolatin1: invalid UTF-8 byte sequence";
        }
    }

private:

    void append(unsigned char c) {
        out.push_back(c);
    }

    void append(char const *s) {
        char c;
        while ((c = *s++) != 0) {
            out.push_back((unsigned char) c);
        }
    }

    bool convert_seq()  {
        unsigned char const c = *ptr++;
        if (c == 0) {
            return false;
        }
        if (c < 128) {
            append(c);
            return true;
        }
        if (c < 0xC0) {
            return utf8Error();
        }
        int count = 1;
        if (c >= 0xE0) {
            ++count;
            if (c >= 0xF0) {
                ++count;
                if (c >=0xF8) {
                    return utf8Error();
                }
            }
        }
        unsigned int mask = 0xFF >> (count + 2);
        unsigned int r = c & mask;
        while (count-- > 0) {
            unsigned char c2 = *ptr++;
            if (c2 == 0 || (c2 & 0xC0) != 0x80) {
                return utf8Error();
            }
            r = r << 6;
            r |= c2 & 0x3F;
        }
        if (r < 256) {
            append((unsigned char) r);
        } else {
            if (r != 0xFEFF) {
                // ignore BOM
                return latin1Error(r);
            }
        }
        return true;
    }

    bool utf8Error() {
        --ptr;
        if (rec == 1 || rec == 2) {
            append('?');
        } else {
            valid = false;
            invalidUnicode = -1;
            return false;
        }
        for (;;) {
            unsigned char const c = *ptr;
            if (c == 0) {
                return false;
            }
            if (c < 0x80 || (c >= 0xC0 && c < 0xF8)) {
                return true;
            }
            ++ptr;
        }
    }

    bool latin1Error(int c) {
        if (rec == 1) {
            append('?');
        } else if (rec == 2) {
            char charref[16];
            sprintf(charref, "&#%d;", c);
            append(charref);
        } else {
            valid = false;
            invalidUnicode = c;
            return false;
        }
        return true;
    }

    std::vector<unsigned char> out;  
    unsigned char const *ptr;
    int rec;
    bool valid;
    int invalidUnicode;
};
