/*
BOIL (Brunni's Own Interpreter Language)
(c) 1998 netEstate (http://www.netestate.de/, info@netestate.de),
Michael Brunnbauer & Markus Hendel, Gesellschaft fuer Netzwerkkommunikation bR
Am Bluetenanger 8, 80995 Muenchen, Germany

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "general.h"
#ifdef WITH_RPC
#include "rpc.h"
#endif
#include "handles.cc"
#include "boil.h"
#include "boil.tab.h"
#include "libs.h"

string strtok_string;
char* strtok_ptr=0;

int args_defined[]= { TYPE_VALUE_STRING,
                      0 };

flowctrl func_defined(type_block* aktblock,opt* options,
                      joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=defined((*args)[0].stringval);
  return back;
}

int args_str_repl[]= { TYPE_VALUE_STRING,
		       TYPE_VALUE_STRING,
		       TYPE_VALUE_STRING,
		       0 };

flowctrl func_str_repl(type_block* aktblock,opt* options,
		       joined_list<type_value>* args)
{
  // FIXME: optimieren!
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval) ||
      !defined((*args)[1].stringval) ||
      !defined((*args)[2].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","str_repl: string not defined");
      return back;
    };

  char* rueck=str_ers((*args)[0].stringval.c_str(),
		      (*args)[1].stringval.c_str(),
		      (*args)[2].stringval.c_str());
  back.returnval.stringval=rueck;
  free(rueck);
  return back;
}

int args_str_repla[]= { TYPE_VALUE_STRING,
			TYPE_VALUE_STRING,
			TYPE_VALUE_STRING,
			0 };

flowctrl func_str_repla(type_block* aktblock,opt* options,
			joined_list<type_value>* args)
{
  // FIXME: optimieren!
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval) ||
      !defined((*args)[1].stringval) ||
      !defined((*args)[2].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","str_repl: string not defined");
      return back;
    };

  char* rueck=str_ersa((*args)[0].stringval.c_str(),
		       (*args)[1].stringval.c_str(),
		       (*args)[2].stringval.c_str());
  back.returnval.stringval=rueck;
  free(rueck);
  return back;
}

int args_strstr[]= { TYPE_VALUE_STRING,
                     TYPE_VALUE_STRING,
                     0 };

flowctrl func_strstr(type_block* aktblock,opt* options,
                     joined_list<type_value>* args)
{
  // FIXME: optimieren!
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval) || !defined((*args)[1].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","strstr: string not defined");
      return back;
    };

  back.returnval.stringval=strstr((*args)[0].stringval.c_str(),(*args)[1].stringval.c_str());

  return back;
}

int args_strtok[]= { TYPE_VALUE_STRING,
                     TYPE_VALUE_STRING,
                     0 };

flowctrl func_strtok(type_block* aktblock,opt* options,
                     joined_list<type_value>* args)
{
  // FIXME: optimieren!
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval) || !defined((*args)[1].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","strtok: string not defined");
      return back;
    };

  strtok_string=(*args)[0].stringval;
  strtok_ptr=strtok_string.lock();
  
  unsigned long i=0;

  if (!strtok_ptr[i])
    {
      setvarulong("error",1);
      setvarstring("perror","strtok: end of string");
      return back;
    };

  while (strtok_ptr[i])
    {
      if (strchr((*args)[1].stringval.c_str(),strtok_ptr[i]))
        {
          strtok_ptr[i++]=0;
          break;
        };
      i++;
    };
  back.returnval.stringval=strtok_ptr;
  strtok_ptr+=i;
  return back;
}

int args_token[]= { TYPE_VALUE_STRING,
                    0 };

flowctrl func_token(type_block* aktblock,opt* options,
                    joined_list<type_value>* args)
{
  // FIXME: optimieren!
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","token: string not defined");
      return back;
    };

  if (!strtok_ptr)
    {
      setvarulong("error",1);
      setvarstring("perror","token: strtok not called");
      return back;
    };

  unsigned long i=0;

  if (!strtok_ptr[i])
    {
      setvarulong("error",1);
      setvarstring("perror","token: end of string");
      return back;
    };

  while (strtok_ptr[i])
    {
      if (strchr((*args)[0].stringval.c_str(),strtok_ptr[i]))
        {
          strtok_ptr[i++]=0;
          break;
        };
      i++;
    };
  back.returnval.stringval=strtok_ptr;
  strtok_ptr+=i;
  return back;
}

int args_strtok1[]= { TYPE_VALUE_STRING,
		      TYPE_VALUE_STRING,
		      0 };

flowctrl func_strtok1(type_block* aktblock,opt* options,
		      joined_list<type_value>* args)
{
  // FIXME: optimieren!
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval) || !defined((*args)[1].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","strtok1: string not defined");
      return back;
    };

  type_value* tokens;
  flowctrl call;
  call=getvar(aktblock,options,tokens,(*args)[0].stringval);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","strtok1: cannot get variable");
      return back;
    };
  if (!defined(tokens->stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","strtok1: token variable undefined");
      return back;
    };
  
  char* ptr=tokens->stringval.lock();
  unsigned long i=0;
  if (!ptr[i])
    {
      setvarulong("error",1);
      setvarstring("perror","strtok1: end of string");
      return back;
    };
  while (ptr[i])
    {
      if (strchr((*args)[1].stringval.c_str(),ptr[i]))
        {
          ptr[i++]=0;
          break;
        };
      i++;
    };
  back.returnval.stringval=ptr;
  string newtokens=ptr+i;
  tokens->stringval=newtokens;
  return back;
}

int args_strtok_esc[]= { TYPE_VALUE_STRING,
                         TYPE_VALUE_STRING,
                         0 };

flowctrl func_strtok_esc(type_block* aktblock,opt* options,
                         joined_list<type_value>* args)
{
  // FIXME: optimieren!
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval) || !defined((*args)[1].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","strtok_esc: string not defined");
      return back;
    };

  strtok_string=(*args)[0].stringval;
  strtok_ptr=strtok_string.lock();
  
  unsigned long i=0;

  if (!strtok_ptr[i])
    {
      setvarulong("error",1);
      setvarstring("perror","strtok_esc: end of string");
      return back;
    };

  while (strtok_ptr[i])
    {
      if (strtok_ptr[i]=='\\')
        { 
          i++;
          if (!strtok_ptr[i])
            {
              setvarulong("error",1);
              setvarstring("perror","strtok: EOF after escape");
              return back;
            };
        }
      else if (strchr((*args)[1].stringval.c_str(),strtok_ptr[i]))
        {
          strtok_ptr[i++]=0;
          break;
        };
      i++;
    };
  back.returnval.stringval=strtok_ptr;
  strtok_ptr+=i;
  return back;
}

int args_token_esc[]= { TYPE_VALUE_STRING,
                        0 };

flowctrl func_token_esc(type_block* aktblock,opt* options,
                        joined_list<type_value>* args)
{
  // FIXME: optimieren!
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","token_esc: string not defined");
      return back;
    };

  unsigned long i=0;

  if (!strtok_ptr[i])
    {
      setvarulong("error",1);
      setvarstring("perror","token_esc: end of string");
      return back;
    };

  while (strtok_ptr[i])
    {
      if (strtok_ptr[i]=='\\')
        {
          i++;
          if (!strtok_ptr[i])
            {
              setvarulong("error",1);
              setvarstring("perror","token_esc: EOF after escape");
              return back;
            };
        }
      else if (strchr((*args)[0].stringval.c_str(),strtok_ptr[i]))
        {
          strtok_ptr[i++]=0;
          break;
        };
      i++;
    };
  back.returnval.stringval=strtok_ptr;
  strtok_ptr+=i;
  return back;
}

int args_strtolower[]= { TYPE_VALUE_STRING,
                         0 };

flowctrl func_strtolower(type_block* aktblock,opt* options,
                         joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","strtolower: string not defined");
      return back;
    };

  back.returnval.stringval=(*args)[0].stringval;
  str_lwr(back.returnval.stringval.lock());

  return back;
}

int args_strtoupper[]= { TYPE_VALUE_STRING,
                         0 };

flowctrl func_strtoupper(type_block* aktblock,opt* options,
                         joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","strtoupper: string not defined");
      return back;
    };

  back.returnval.stringval=(*args)[0].stringval;
  str_upr(back.returnval.stringval.lock());

  return back;
}

int args_substr[]= { TYPE_VALUE_STRING,
                     TYPE_VALUE_ULONG,
                     TYPE_VALUE_ULONG,
                     0 };

flowctrl func_substr(type_block* aktblock,opt* options,
                     joined_list<type_value>* args)
{
  // FIXME: optimieren!
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","substr: string not defined");
      return back;
    };

  if ((*args)[1].val.ulongval > (*args)[0].stringval.size())
    {
      setvarulong("error",1);
      setvarstring("perror","substr: position to high");
      return back;
    };

  back.returnval.stringval=(*args)[0].stringval.c_str()+(*args)[1].val.ulongval;

  if ((*args)[2].val.ulongval)
    {
      if ((*args)[2].val.ulongval > back.returnval.stringval.size())
        {
          setvarulong("error",1);
          setvarstring("perror","substr: length to high");
          return back;
        };
      
      back.returnval.stringval.lock()[(*args)[2].val.ulongval]=0;
      back.returnval.stringval.unlock();
    };

  return back;
}

int args_strlen[]= { TYPE_VALUE_STRING,
                     0 };

flowctrl func_strlen(type_block* aktblock,opt* options,
                     joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","strlen: string not defined");
      return back;
    };

  back.returnval.val.ulongval=(*args)[0].stringval.size();

  return back;
}

int args_escape[]= { TYPE_VALUE_STRING,
                     TYPE_VALUE_STRING,
                     0 };

flowctrl func_escape(type_block* aktblock,opt* options,
                     joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  char buf[2];
  buf[1]=0;

  if (!defined((*args)[0].stringval) || !defined((*args)[1].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","escape: string not defined");
      return back;
    };

  back.returnval.stringval="";
  unsigned long len=(*args)[0].stringval.size();
  unsigned long len1=(*args)[1].stringval.size();
  for (unsigned long i=0;i!=len;i++)
    {
      buf[0]=(*args)[0].stringval[i];
      if (buf[0]=='\\')
        {
          back.returnval.stringval+="\\";
        }
      else
        {
          for (unsigned long u=0;u!=len1;u++)
            {
              if (buf[0]==(*args)[1].stringval[u])
                {
                  back.returnval.stringval+="\\";
                  break;
                };
            };
        };
      back.returnval.stringval+=buf;
    };

  return back;
}

int args_unescape[]= { TYPE_VALUE_STRING,
                       0 };

flowctrl func_unescape(type_block* aktblock,opt* options,
                       joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  char buf[2];
  buf[1]=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","unescape: string not defined");
      return back;
    };

  back.returnval.stringval="";
  unsigned long len=(*args)[0].stringval.size();
  for (unsigned long i=0;i!=len;i++)
    {
      buf[0]=(*args)[0].stringval[i];
      if (buf[0]!='\\')
        {
          back.returnval.stringval+=buf;
        }
      else if (i+1 < len && (*args)[0].stringval[i+1]=='\\')
        {
          back.returnval.stringval+="\\";
	  i++;
        };
    };

  return back;
}

int args_escape1[]= { TYPE_VALUE_STRING,
                      0 };

flowctrl func_escape1(type_block* aktblock,opt* options,
                      joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  char buf[2];
  buf[1]=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","escape1: string not defined");
      return back;
    };

  back.returnval.stringval="";
  unsigned long len=(*args)[0].stringval.size();
  for (unsigned long i=0;i!=len;i++)
    {
      buf[0]=(*args)[0].stringval[i];
      if (buf[0]=='\\') { back.returnval.stringval+="\\"; }
      else if (buf[0]=='\n')
        { buf[0]='n'; back.returnval.stringval+="\\"; }
      else if (buf[0]=='\r')
        { buf[0]='n'; back.returnval.stringval+="\\"; }
      else if (buf[0]=='\v')
        { buf[0]='v'; back.returnval.stringval+="\\"; }
      else if (buf[0]=='\f')
        { buf[0]='f'; back.returnval.stringval+="\\"; };
            back.returnval.stringval+=buf;
    };

  return back;
}

int args_unescape1[]= { TYPE_VALUE_STRING,
                        0 };

flowctrl func_unescape1(type_block* aktblock,opt* options,
                        joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  char buf[2];
  buf[1]=0;
  
  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","unescape: string not defined");
      return back;
    };

  back.returnval.stringval="";
  unsigned long len=(*args)[0].stringval.size();
  for (unsigned long i=0;i<len;i++)

    {
      buf[0]=(*args)[0].stringval[i];
      if (buf[0]!='\\')
        {
          back.returnval.stringval+=buf;
        }
      else if (i+1 < len)
        {
          if ((*args)[0].stringval[i+1]=='\\')
            {
              back.returnval.stringval+="\\";
              i++;
            }
          else if ((*args)[0].stringval[i+1]=='n')
            {
              back.returnval.stringval+="\n";
              i++;
            }
          else if ((*args)[0].stringval[i+1]=='r')
            {
              back.returnval.stringval+="\r";
              i++;
            }
          else if ((*args)[0].stringval[i+1]=='v')
            {
              back.returnval.stringval+="\v";
              i++;
            }
          else if ((*args)[0].stringval[i+1]=='f')
            {
              back.returnval.stringval+="\f";
              i++;
            };
        };
    };

  return back;
}

int args_textfile[]= { TYPE_VALUE_STRING,
                       0 };

flowctrl func_textfile(type_block* aktblock,opt* options,
                       joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  char buf[MAX_LINE_SIZE];
  buf[MAX_LINE_SIZE-1]=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","textfile: string not defined");
      return back;
    };

  int handle=open((*args)[0].stringval.c_str(),O_RDONLY);
  if (handle==-1)
    { 
      setvarulong("error",1);
      setvarstring("perror","textfile: cannot open "+(*args)[0].stringval);
      return back;
    };

  back.returnval.stringval="";
  ssize_t size;
  while ((size=read(handle,buf,MAX_LINE_SIZE-1))>=0)
    {
      if (!size)
	{
	  if (errno==EINTR) { errno=0; } else { break; };
	};
      buf[size]=0;
      back.returnval.stringval+=buf;
    };
  close(handle);

  return back;
}

int args_getchar[]= { TYPE_VALUE_STRING,
                      TYPE_VALUE_ULONG,
                      0 };

flowctrl func_getchar(type_block* aktblock,opt* options,
                      joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","getchar: string not defined");
      return back;
    };

  if ((*args)[1].val.ulongval>=(*args)[0].stringval.size())
    {
      setvarulong("error",1);
      setvarstring("perror","getchar: index too big");
      return back;
    };

  unsigned char ccc=(*args)[0].stringval[(*args)[1].val.ulongval];
  back.returnval.val.ulongval=(unsigned long) ccc;
  return back;
}

int args_sql_escape[]= { TYPE_VALUE_STRING,
                         0 };

flowctrl func_sql_escape(type_block* aktblock,opt* options,
                         joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  char buf[2];
  buf[1]=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","sql_escape: string not defined");
      return back;
    };

  back.returnval.stringval="";
  unsigned long len=(*args)[0].stringval.size();
  for (unsigned long i=0;i!=len;i++)
    {
      buf[0]=(*args)[0].stringval[i];
      if (buf[0]=='\'' || buf[0]=='"' || buf[0]=='\\')
        {
          back.returnval.stringval+="\\";
        };
      back.returnval.stringval+=buf;
    };

  return back;
}

typedef unsigned long uint32;

struct MD5Context {
        uint32 buf[4];
        uint32 bits[2];
        unsigned char in[64];
};

/*
 * Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
 * initialization constants.
 */
void MD5Init(struct MD5Context *ctx)
{
    ctx->buf[0] = 0x67452301;
    ctx->buf[1] = 0xefcdab89;
    ctx->buf[2] = 0x98badcfe;
    ctx->buf[3] = 0x10325476;

    ctx->bits[0] = 0;
    ctx->bits[1] = 0;
}

#ifndef ASM_MD5

/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#ifdef __PUREC__
#define MD5STEP(f, w, x, y, z, data, s) \
( w += f /*(x, y, z)*/ + data,  w = w<<s | w>>(32-s),  w += x )
#else
#define MD5STEP(f, w, x, y, z, data, s) \
  ( w += f(x, y, z) + data,  w = w<<s | w>>(32-s),  w += x )
#endif

/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
void MD5Transform(uint32 buf[4], uint32 const in[16])
{
    register uint32 a, b, c, d;

    a = buf[0];
    b = buf[1];
    c = buf[2];
    d = buf[3];

#ifdef __PUREC__        /* PureC Weirdness... (GG) */
    MD5STEP(F1(b,c,d), a, b, c, d, in[0] + 0xd76aa478L, 7);
    MD5STEP(F1(a,b,c), d, a, b, c, in[1] + 0xe8c7b756L, 12);
    MD5STEP(F1(d,a,b), c, d, a, b, in[2] + 0x242070dbL, 17);
    MD5STEP(F1(c,d,a), b, c, d, a, in[3] + 0xc1bdceeeL, 22);
    MD5STEP(F1(b,c,d), a, b, c, d, in[4] + 0xf57c0fafL, 7);
    MD5STEP(F1(a,b,c), d, a, b, c, in[5] + 0x4787c62aL, 12);
    MD5STEP(F1(d,a,b), c, d, a, b, in[6] + 0xa8304613L, 17);
    MD5STEP(F1(c,d,a), b, c, d, a, in[7] + 0xfd469501L, 22);
    MD5STEP(F1(b,c,d), a, b, c, d, in[8] + 0x698098d8L, 7);
    MD5STEP(F1(a,b,c), d, a, b, c, in[9] + 0x8b44f7afL, 12);
    MD5STEP(F1(d,a,b), c, d, a, b, in[10] + 0xffff5bb1L, 17);
    MD5STEP(F1(c,d,a), b, c, d, a, in[11] + 0x895cd7beL, 22);
    MD5STEP(F1(b,c,d), a, b, c, d, in[12] + 0x6b901122L, 7);
    MD5STEP(F1(a,b,c), d, a, b, c, in[13] + 0xfd987193L, 12);
    MD5STEP(F1(d,a,b), c, d, a, b, in[14] + 0xa679438eL, 17);
    MD5STEP(F1(c,d,a), b, c, d, a, in[15] + 0x49b40821L, 22);

    MD5STEP(F2(b,c,d), a, b, c, d, in[1] + 0xf61e2562L, 5);
    MD5STEP(F2(a,b,c), d, a, b, c, in[6] + 0xc040b340L, 9);
    MD5STEP(F2(d,a,b), c, d, a, b, in[11] + 0x265e5a51L, 14);
    MD5STEP(F2(c,d,a), b, c, d, a, in[0] + 0xe9b6c7aaL, 20);
    MD5STEP(F2(b,c,d), a, b, c, d, in[5] + 0xd62f105dL, 5);
    MD5STEP(F2(a,b,c), d, a, b, c, in[10] + 0x02441453L, 9);
    MD5STEP(F2(d,a,b), c, d, a, b, in[15] + 0xd8a1e681L, 14);
    MD5STEP(F2(c,d,a), b, c, d, a, in[4] + 0xe7d3fbc8L, 20);
    MD5STEP(F2(b,c,d), a, b, c, d, in[9] + 0x21e1cde6L, 5);
    MD5STEP(F2(a,b,c), d, a, b, c, in[14] + 0xc33707d6L, 9);
    MD5STEP(F2(d,a,b), c, d, a, b, in[3] + 0xf4d50d87L, 14);
    MD5STEP(F2(c,d,a), b, c, d, a, in[8] + 0x455a14edL, 20);
    MD5STEP(F2(b,c,d), a, b, c, d, in[13] + 0xa9e3e905L, 5);
    MD5STEP(F2(a,b,c), d, a, b, c, in[2] + 0xfcefa3f8L, 9);
    MD5STEP(F2(d,a,b), c, d, a, b, in[7] + 0x676f02d9L, 14);
    MD5STEP(F2(c,d,a), b, c, d, a, in[12] + 0x8d2a4c8aL, 20);

    MD5STEP(F3(b,c,d), a, b, c, d, in[5] + 0xfffa3942L, 4);
    MD5STEP(F3(a,b,c), d, a, b, c, in[8] + 0x8771f681L, 11);
    MD5STEP(F3(d,a,b), c, d, a, b, in[11] + 0x6d9d6122L, 16);
    MD5STEP(F3(c,d,a), b, c, d, a, in[14] + 0xfde5380cL, 23);
    MD5STEP(F3(b,c,d), a, b, c, d, in[1] + 0xa4beea44L, 4);
    MD5STEP(F3(a,b,c), d, a, b, c, in[4] + 0x4bdecfa9L, 11);
    MD5STEP(F3(d,a,b), c, d, a, b, in[7] + 0xf6bb4b60L, 16);
    MD5STEP(F3(c,d,a), b, c, d, a, in[10] + 0xbebfbc70L, 23);
    MD5STEP(F3(b,c,d), a, b, c, d, in[13] + 0x289b7ec6L, 4);
    MD5STEP(F3(a,b,c), d, a, b, c, in[0] + 0xeaa127faL, 11);
    MD5STEP(F3(d,a,b), c, d, a, b, in[3] + 0xd4ef3085L, 16);
    MD5STEP(F3(c,d,a), b, c, d, a, in[6] + 0x04881d05L, 23);
    MD5STEP(F3(b,c,d), a, b, c, d, in[9] + 0xd9d4d039L, 4);
    MD5STEP(F3(a,b,c), d, a, b, c, in[12] + 0xe6db99e5L, 11);
    MD5STEP(F3(d,a,b), c, d, a, b, in[15] + 0x1fa27cf8L, 16);
    MD5STEP(F3(c,d,a), b, c, d, a, in[2] + 0xc4ac5665L, 23);

    MD5STEP(F4(b,c,d), a, b, c, d, in[0] + 0xf4292244L, 6);
    MD5STEP(F4(a,b,c), d, a, b, c, in[7] + 0x432aff97L, 10);
    MD5STEP(F4(d,a,b), c, d, a, b, in[14] + 0xab9423a7L, 15);
    MD5STEP(F4(c,d,a), b, c, d, a, in[5] + 0xfc93a039L, 21);
    MD5STEP(F4(b,c,d), a, b, c, d, in[12] + 0x655b59c3L, 6);
    MD5STEP(F4(a,b,c), d, a, b, c, in[3] + 0x8f0ccc92L, 10);
    MD5STEP(F4(d,a,b), c, d, a, b, in[10] + 0xffeff47dL, 15);
    MD5STEP(F4(c,d,a), b, c, d, a, in[1] + 0x85845dd1L, 21);
    MD5STEP(F4(b,c,d), a, b, c, d, in[8] + 0x6fa87e4fL, 6);
    MD5STEP(F4(a,b,c), d, a, b, c, in[15] + 0xfe2ce6e0L, 10);
    MD5STEP(F4(d,a,b), c, d, a, b, in[6] + 0xa3014314L, 15);
    MD5STEP(F4(c,d,a), b, c, d, a, in[13] + 0x4e0811a1L, 21);
    MD5STEP(F4(b,c,d), a, b, c, d, in[4] + 0xf7537e82L, 6);
    MD5STEP(F4(a,b,c), d, a, b, c, in[11] + 0xbd3af235L, 10);
    MD5STEP(F4(d,a,b), c, d, a, b, in[2] + 0x2ad7d2bbL, 15);
    MD5STEP(F4(c,d,a), b, c, d, a, in[9] + 0xeb86d391L, 21);
#else
    MD5STEP(F1, a, b, c, d, in[0] + 0xd76aa478, 7);
    MD5STEP(F1, d, a, b, c, in[1] + 0xe8c7b756, 12);
    MD5STEP(F1, c, d, a, b, in[2] + 0x242070db, 17);
    MD5STEP(F1, b, c, d, a, in[3] + 0xc1bdceee, 22);
    MD5STEP(F1, a, b, c, d, in[4] + 0xf57c0faf, 7);
    MD5STEP(F1, d, a, b, c, in[5] + 0x4787c62a, 12);
    MD5STEP(F1, c, d, a, b, in[6] + 0xa8304613, 17);
    MD5STEP(F1, b, c, d, a, in[7] + 0xfd469501, 22);
    MD5STEP(F1, a, b, c, d, in[8] + 0x698098d8, 7);
    MD5STEP(F1, d, a, b, c, in[9] + 0x8b44f7af, 12);
    MD5STEP(F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
    MD5STEP(F1, b, c, d, a, in[11] + 0x895cd7be, 22);
    MD5STEP(F1, a, b, c, d, in[12] + 0x6b901122, 7);
    MD5STEP(F1, d, a, b, c, in[13] + 0xfd987193, 12);
    MD5STEP(F1, c, d, a, b, in[14] + 0xa679438e, 17);
    MD5STEP(F1, b, c, d, a, in[15] + 0x49b40821, 22);

    MD5STEP(F2, a, b, c, d, in[1] + 0xf61e2562, 5);
    MD5STEP(F2, d, a, b, c, in[6] + 0xc040b340, 9);
    MD5STEP(F2, c, d, a, b, in[11] + 0x265e5a51, 14);
    MD5STEP(F2, b, c, d, a, in[0] + 0xe9b6c7aa, 20);
    MD5STEP(F2, a, b, c, d, in[5] + 0xd62f105d, 5);
    MD5STEP(F2, d, a, b, c, in[10] + 0x02441453, 9);
    MD5STEP(F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
    MD5STEP(F2, b, c, d, a, in[4] + 0xe7d3fbc8, 20);
    MD5STEP(F2, a, b, c, d, in[9] + 0x21e1cde6, 5);
    MD5STEP(F2, d, a, b, c, in[14] + 0xc33707d6, 9);
    MD5STEP(F2, c, d, a, b, in[3] + 0xf4d50d87, 14);
    MD5STEP(F2, b, c, d, a, in[8] + 0x455a14ed, 20);
    MD5STEP(F2, a, b, c, d, in[13] + 0xa9e3e905, 5);
    MD5STEP(F2, d, a, b, c, in[2] + 0xfcefa3f8, 9);
    MD5STEP(F2, c, d, a, b, in[7] + 0x676f02d9, 14);
    MD5STEP(F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

    MD5STEP(F3, a, b, c, d, in[5] + 0xfffa3942, 4);
    MD5STEP(F3, d, a, b, c, in[8] + 0x8771f681, 11);
    MD5STEP(F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
    MD5STEP(F3, b, c, d, a, in[14] + 0xfde5380c, 23);
    MD5STEP(F3, a, b, c, d, in[1] + 0xa4beea44, 4);
    MD5STEP(F3, d, a, b, c, in[4] + 0x4bdecfa9, 11);
    MD5STEP(F3, c, d, a, b, in[7] + 0xf6bb4b60, 16);
    MD5STEP(F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
    MD5STEP(F3, a, b, c, d, in[13] + 0x289b7ec6, 4);
    MD5STEP(F3, d, a, b, c, in[0] + 0xeaa127fa, 11);
    MD5STEP(F3, c, d, a, b, in[3] + 0xd4ef3085, 16);
    MD5STEP(F3, b, c, d, a, in[6] + 0x04881d05, 23);
    MD5STEP(F3, a, b, c, d, in[9] + 0xd9d4d039, 4);
    MD5STEP(F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
    MD5STEP(F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
    MD5STEP(F3, b, c, d, a, in[2] + 0xc4ac5665, 23);

    MD5STEP(F4, a, b, c, d, in[0] + 0xf4292244, 6);
    MD5STEP(F4, d, a, b, c, in[7] + 0x432aff97, 10);
    MD5STEP(F4, c, d, a, b, in[14] + 0xab9423a7, 15);
    MD5STEP(F4, b, c, d, a, in[5] + 0xfc93a039, 21);
    MD5STEP(F4, a, b, c, d, in[12] + 0x655b59c3, 6);
    MD5STEP(F4, d, a, b, c, in[3] + 0x8f0ccc92, 10);
    MD5STEP(F4, c, d, a, b, in[10] + 0xffeff47d, 15);
    MD5STEP(F4, b, c, d, a, in[1] + 0x85845dd1, 21);
    MD5STEP(F4, a, b, c, d, in[8] + 0x6fa87e4f, 6);
    MD5STEP(F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
    MD5STEP(F4, c, d, a, b, in[6] + 0xa3014314, 15);
    MD5STEP(F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
    MD5STEP(F4, a, b, c, d, in[4] + 0xf7537e82, 6);
    MD5STEP(F4, d, a, b, c, in[11] + 0xbd3af235, 10);
    MD5STEP(F4, c, d, a, b, in[2] + 0x2ad7d2bb, 15);
    MD5STEP(F4, b, c, d, a, in[9] + 0xeb86d391, 21);
#endif

    buf[0] += a;
    buf[1] += b;
    buf[2] += c;
    buf[3] += d;
    return;
}

#endif

/*
 * Update context to reflect the concatenation of another buffer full
 * of bytes.
 */
void MD5Update(struct MD5Context *ctx, unsigned char const *buf, unsigned len)
{
    uint32 t;

    /* Update bitcount */

    t = ctx->bits[0];
    if ((ctx->bits[0] = t + ((uint32) len << 3)) < t)
        ctx->bits[1]++;         /* Carry from low to high */
    ctx->bits[1] += len >> 29;

    t = (t >> 3) & 0x3f;        /* Bytes already in shsInfo->data */

    /* Handle any leading odd-sized chunks */

    if (t) {
        unsigned char *p = (unsigned char *) ctx->in + t;

        t = 64 - t;
        if (len < t) {
            memcpy(p, buf, len);
            return;
        }
        memcpy(p, buf, t);
        MD5Transform(ctx->buf, (uint32 *) ctx->in);
        buf += t;
        len -= t;
    }
    /* Process data in 64-byte chunks */

    while (len >= 64) {
        memcpy(ctx->in, buf, 64);
        MD5Transform(ctx->buf, (uint32 *) ctx->in);
        buf += 64;
        len -= 64;
    }

    /* Handle any remaining bytes of data. */

    memcpy(ctx->in, buf, len);
}

/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern
 * 1 0* (64-bit count of bits processed, MSB-first)
 */
void MD5Final(unsigned char digest[16], struct MD5Context *ctx)
{
    unsigned count;
    unsigned char *p;

    /* Compute number of bytes mod 64 */
    count = (ctx->bits[0] >> 3) & 0x3F;

    /* Set the first char of padding to 0x80.  This is safe since there is
       always at least one byte free */
    p = ctx->in + count;
    *p++ = 0x80;

    /* Bytes of padding needed to make 64 bytes */
    count = 64 - 1 - count;

    /* Pad out to 56 mod 64 */
    if (count < 8) {
      /* Two lots of padding:  Pad the first block to 64 bytes */
        memset(p, 0, count);
        MD5Transform(ctx->buf, (uint32 *) ctx->in);

        /* Now fill the next block with 56 bytes */
        memset(ctx->in, 0, 56);
    } else {
      /* Pad block to 56 bytes */
        memset(p, 0, count - 8);
    }

    /* Append length in bits and transform */
    ((uint32 *) ctx->in)[14] = ctx->bits[0];
    ((uint32 *) ctx->in)[15] = ctx->bits[1];

    MD5Transform(ctx->buf, (uint32 *) ctx->in);
    memcpy(digest, ctx->buf, 16);
    memset(ctx, 0, sizeof(ctx));        /* In case it's sensitive */
}

int args_md5[]= { TYPE_VALUE_STRING,
		  0 };

flowctrl func_md5(type_block* aktblock,opt* options,
		  joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","md5: string not defined");
      return back;
    };

  struct MD5Context ctx;
  const char* buf=(*args)[0].stringval.c_str();
  unsigned char digest[16];
  char digest1[33];
  int i;

  MD5Init(&ctx);
  MD5Update(&ctx, (unsigned char*) buf, strlen(buf));
  MD5Final(digest, &ctx);
  
  for (i = 0; i < 16; i++) sprintf(digest1+2*i,"%02x",digest[i]);
  digest1[32]=0;
  
  back.returnval.stringval=digest1;
  return back;
}

/*********************************************************************\

MODULE NAME:    b64.c

AUTHOR:         Bob Trower 08/04/01

PROJECT:        Crypt Data Packaging

COPYRIGHT:      Copyright (c) Trantor Standard Systems Inc., 2001

NOTE:           This source code may be used as you wish, subject to
                the MIT license.  See the LICENCE section below.

DESCRIPTION:
                This little utility implements the Base64
                Content-Transfer-Encoding standard described in
                RFC1113 (http://www.faqs.org/rfcs/rfc1113.html).

                This is the coding scheme used by MIME to allow
                binary data to be transferred by SMTP mail.

                Groups of 3 bytes from a binary stream are coded as
                groups of 4 bytes in a text stream.

                The input stream is 'padded' with zeros to create
                an input that is an even multiple of 3.

                A special character ('=') is used to denote padding so
                that the stream can be decoded back to its exact size.

                Encoded output is formatted in lines which should
                be a maximum of 72 characters to conform to the
                specification.  This program defaults to 72 characters,
                but will allow more or less through the use of a
                switch.  The program enforces a minimum line size
                of 4 characters.

                Example encoding:

                The stream 'ABCD' is 32 bits long.  It is mapped as
                follows:

                ABCD

                 A (65)     B (66)     C (67)     D (68)   (None) (None)
                01000001   01000010   01000011   01000100

                16 (Q)  20 (U)  9 (J)   3 (D)    17 (R) 0 (A)  NA (=) NA (=)
                010000  010100  001001  000011   010001 000000 000000 000000


                QUJDRA==

                Decoding is the process in reverse.  A 'decode' lookup
                table has been created to avoid string scans.

DESIGN GOALS:	Specifically:
		Code is a stand-alone utility to perform base64 
		encoding/decoding. It should be genuinely useful 
		when the need arises and it meets a need that is 
		likely to occur for some users.  
		Code acts as sample code to show the author's 
		design and coding style.  

		Generally: 
		This program is designed to survive:
		Everything you need is in a single source file.
		It compiles cleanly using a vanilla ANSI C compiler.
		It does its job correctly with a minimum of fuss.  
		The code is not overly clever, not overly simplistic 
		and not overly verbose. 
		Access is 'cut and paste' from a web page.  
		Terms of use are reasonable.  

VALIDATION:     Non-trivial code is never without errors.  This
                file likely has some problems, since it has only
                been tested by the author.  It is expected with most
                source code that there is a period of 'burn-in' when
                problems are identified and corrected.  That being
                said, it is possible to have 'reasonably correct'
                code by following a regime of unit test that covers
                the most likely cases and regression testing prior
                to release.  This has been done with this code and
                it has a good probability of performing as expected.

                Unit Test Cases:

                case 0:empty file:
                    CASE0.DAT  ->  ->
                    (Zero length target file created
                    on both encode and decode.)

                case 1:One input character:
                    CASE1.DAT A -> QQ== -> A

                case 2:Two input characters:
                    CASE2.DAT AB -> QUJD -> AB

                case 3:Three input characters:
                    CASE3.DAT ABC -> QUJD -> ABC

                case 4:Four input characters:
                    case4.dat ABCD -> QUJDRA== -> ABCD

                case 5:All chars from 0 to ff, linesize set to 50:

                    AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIj
                    JCUmJygpKissLS4vMDEyMzQ1Njc4OTo7PD0+P0BBQkNERUZH
                    SElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWpr
                    bG1ub3BxcnN0dXZ3eHl6e3x9fn+AgYKDhIWGh4iJiouMjY6P
                    kJGSk5SVlpeYmZqbnJ2en6ChoqOkpaanqKmqq6ytrq+wsbKz
                    tLW2t7i5uru8vb6/wMHCw8TFxsfIycrLzM3Oz9DR0tPU1dbX
                    2Nna29zd3t/g4eLj5OXm5+jp6uvs7e7v8PHy8/T19vf4+fr7
                    /P3+/w==

                case 6:Mime Block from e-mail:
                    (Data same as test case 5)

                case 7: Large files:
                    Tested 28 MB file in/out.

                case 8: Random Binary Integrity:
                    This binary program (b64.exe) was encoded to base64,
                    back to binary and then executed.

                case 9 Stress:
                    All files in a working directory encoded/decoded
                    and compared with file comparison utility to
                    ensure that multiple runs do not cause problems
                    such as exhausting file handles, tmp storage, etc.

                -------------

                Syntax, operation and failure:
                    All options/switches tested.  Performs as
                    expected.

                case 10:
                    No Args -- Shows Usage Screen
                    Return Code 1 (Invalid Syntax)
                case 11:
                    One Arg (invalid) -- Shows Usage Screen
                    Return Code 1 (Invalid Syntax)
                case 12:
                    One Arg Help (-?) -- Shows detailed Usage Screen.
                    Return Code 0 (Success -- help request is valid).
                case 13:
                    One Arg Help (-h) -- Shows detailed Usage Screen.
                    Return Code 0 (Success -- help request is valid).
                case 14:
                    One Arg (valid) -- Uses stdin/stdout (filter)
                    Return Code 0 (Sucess)
                case 15:
                    Two Args (invalid file) -- shows system error.
                    Return Code 2 (File Error)
                case 16:
                    Encode non-existent file -- shows system error.
                    Return Code 2 (File Error)
                case 17:
                    Out of disk space -- shows system error.
                    Return Code 3 (File I/O Error)

                -------------

                Compile/Regression test:
                    gcc compiled binary under Cygwin
                    Microsoft Visual Studio under Windows 2000
                    Microsoft Version 6.0 C under Windows 2000

DEPENDENCIES:   None

LICENCE:        Copyright (c) 2001 Bob Trower, Trantor Standard Systems Inc.

                Permission is hereby granted, free of charge, to any person
                obtaining a copy of this software and associated
                documentation files (the "Software"), to deal in the
                Software without restriction, including without limitation
                the rights to use, copy, modify, merge, publish, distribute,
                sublicense, and/or sell copies of the Software, and to
                permit persons to whom the Software is furnished to do so,
                subject to the following conditions:

                The above copyright notice and this permission notice shall
                be included in all copies or substantial portions of the
                Software.

                THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
                KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
                WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
                PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
                OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
                OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
                OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
                SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

VERSION HISTORY:
                Bob Trower 08/04/01 -- Create Version 0.00.00B

\******************************************************************* */

static const char cb64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static const char cd64[]="|$$$}rstuvwxyz{$$$$$$$>?@ABCDEFGHIJKLMNOPQRSTUVW$$$$$$XYZ[\\]^_`abcdefghijklmnopq";

void encodeblock( unsigned char in[3], unsigned char out[4], int len )
{
    out[0] = cb64[ in[0] >> 2 ];
    out[1] = cb64[ ((in[0] & 0x03) << 4) | ((in[1] & 0xf0) >> 4) ];
    out[2] = (unsigned char) (len > 1 ? cb64[ ((in[1] & 0x0f) << 2) | ((in[2] & 0xc0) >> 6) ] : '=');
    out[3] = (unsigned char) (len > 2 ? cb64[ in[2] & 0x3f ] : '=');
}

#define B64_DEF_LINE_SIZE   72

char* encode( FILE *infile)
{
    char* outbuf;
    char* outbuf_new;
    unsigned long outlen=0;
    unsigned long outbuflen=65536;

    unsigned char in[3], out[4];
    int i, len, blocksout = 0;

    outbuf=(char*) malloc(outbuflen);
    if (!outbuf) return 0;

    while( !feof( infile ) ) {
        len = 0;
        for( i = 0; i < 3; i++ ) {
            in[i] = (unsigned char) getc( infile );
            if( !feof( infile ) ) {
                len++;
            }
            else {
                in[i] = 0;
            }
        }
        if( len ) {
            encodeblock( in, out, len );
            for( i = 0; i < 4; i++ ) {
	      
	      if (outlen+2>outbuflen)
                {
		  outbuflen=outbuflen*2;
		  outbuf_new=(char*) malloc(outbuflen);
		  if (!outbuf_new)
		    {
		      free(outbuf);
		      return 0;
		    };
		  strncpy(outbuf_new,outbuf,outlen);
		  outbuf=outbuf_new;
                };
	      outbuf[outlen++]=out[i];
            }
            blocksout++;
        }
        if( blocksout >= (B64_DEF_LINE_SIZE/4) || feof( infile ) ) {
            if( blocksout ) {

	      if (outlen+3>outbuflen)
		{
		  outbuflen=outbuflen*2;
		  outbuf_new=(char*) malloc(outbuflen);
		  if (!outbuf_new)
		    {
		      free(outbuf);
		      return 0;
		    };
		  strncpy(outbuf_new,outbuf,outlen);
		  outbuf=outbuf_new;
		};
              outbuf[outlen++]='\r';
              outbuf[outlen++]='\n';
            }
            blocksout = 0;
        }
    }
    outbuf[outlen]=0;
    return outbuf;
}

char* encode_string(char const *instr)
{
    char* outbuf;
    char* outbuf_new;
    unsigned long outlen=0;
    unsigned long outbuflen=65536;
    char const *p = instr;

    unsigned char in[3], out[4];
    int i, len, blocksout = 0;

    outbuf=(char*) malloc(outbuflen);
    if (!outbuf) return 0;

    while(*p) {
        len = 0;
        for( i = 0; i < 3; i++ ) {
            in[i] = (unsigned char) *p;
            if(*p) {
                len++;
                ++p;
            }
            else {
                in[i] = 0;
            }
        }
        if(len) {
            encodeblock( in, out, len );
            for( i = 0; i < 4; i++ ) {
	      
	      if (outlen+2>outbuflen)
                {
		  outbuflen=outbuflen*2;
		  outbuf_new=(char*) malloc(outbuflen);
		  if (!outbuf_new)
		    {
		      free(outbuf);
		      return 0;
		    };
		  strncpy(outbuf_new,outbuf,outlen);
		  outbuf=outbuf_new;
                };
	      outbuf[outlen++]=out[i];
            }
            blocksout++;
        }
        if( blocksout >= (B64_DEF_LINE_SIZE/4) || !*p ) {
            if( blocksout ) {

	      if (outlen+3>outbuflen)
		{
		  outbuflen=outbuflen*2;
		  outbuf_new=(char*) malloc(outbuflen);
		  if (!outbuf_new)
		    {
		      free(outbuf);
		      return 0;
		    };
		  strncpy(outbuf_new,outbuf,outlen);
		  outbuf=outbuf_new;
		};
              outbuf[outlen++]='\r';
              outbuf[outlen++]='\n';
            }
            blocksout = 0;
        }
    }
    outbuf[outlen]=0;
    return outbuf;
}

void decodeblock( unsigned char in[4], unsigned char out[3] )
{   
    out[ 0 ] = (unsigned char ) (in[0] << 2 | in[1] >> 4);
    out[ 1 ] = (unsigned char ) (in[1] << 4 | in[2] >> 2);
    out[ 2 ] = (unsigned char ) (((in[2] << 6) & 0xc0) | in[3]);
}

void decode( char* inbuf, FILE *outfile )
{
    unsigned char in[4], out[3], v;
    int i, len;

    while( *inbuf ) {
        for( len = 0, i = 0; i < 4 && *inbuf; i++ ) {
            v = 0;
            while( *inbuf && v == 0 ) {
                v = (unsigned char) *inbuf;
                inbuf++;
                v = (unsigned char) ((v < 43 || v > 122) ? 0 : cd64[ v - 43 ]);
                if( v ) {
                    v = (unsigned char) ((v == '$') ? 0 : v - 61);
                }
            }
            if( *inbuf ) {
                len++;
                if( v ) {
                    in[ i ] = (unsigned char) (v - 1);
                }
            }
            else {
                in[i] = 0;
            }
        }
        if( len ) {
            decodeblock( in, out );
            for( i = 0; i < len - 1; i++ ) {
                putc( out[i], outfile );
            }
        }
    }
}

int args_base64encode[]= { TYPE_VALUE_LONG,
	                  0 };

flowctrl func_base64encode(type_block* aktblock,opt* options,
			   joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  FILE* fh;

  if (!options->ptr_h->streams.ishandle((*args)[0].val.longval))
    {
      setvarulong("error",1);
      setvarstring("perror","base64encode: illegal handle");
      return back;
    };
  fh=options->ptr_h->streams.gethandle((*args)[0].val.longval);

  char* b64=encode(fh);
  if (!b64)
    {
      setvarulong("error",1);
      setvarstring("perror","base64encode: out of memory");
      return back;
    };
  back.returnval.stringval=b64;
  free(b64);
  return back;
}

int args_base64stringencode[]= { TYPE_VALUE_STRING,
	                         0 };

flowctrl func_base64stringencode(type_block* aktblock,opt* options,
			   joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
  {
      setvarulong("error",1);
      setvarstring("perror","base64stringdecode: string not defined");
      return back;
  }

  char *b64 = encode_string((*args)[0].stringval.c_str());
  if (!b64)
    {
      setvarulong("error",1);
      setvarstring("perror","base64encode: out of memory");
      return back;
    };
  back.returnval.stringval=b64;
  free(b64);
  return back;
}

int args_base64decode[]= { TYPE_VALUE_STRING,
			   TYPE_VALUE_LONG,
			   0 };

flowctrl func_base64decode(type_block* aktblock,opt* options,
                           joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_LONG;
  back.returnval.val.longval=0;

  FILE* fh;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","base64decode: string not defined");
      return back;
    };

  if (!options->ptr_h->streams.ishandle((*args)[1].val.longval))
    {
      setvarulong("error",1);
      setvarstring("perror","base64decode: illegal handle");
      return back;
    };
  fh=options->ptr_h->streams.gethandle((*args)[1].val.longval);

  decode((*args)[0].stringval.lock(),fh);

  back.returnval.val.longval=1;
  return back;
}

#include "Utf8ToLatin1.h"

int args_utf8tolatin1[]= { TYPE_VALUE_STRING,
                           TYPE_VALUE_LONG,
                           0 };

flowctrl func_utf8tolatin1(type_block *aktblock,
                           opt *options,
                           joined_list<type_value> *args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back,call,call1;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror", "utf8tolatin1: string not defined");
      return back;
    };

  Utf8ToLatin1 converter((*args)[0].stringval.c_str(),
                         (*args)[1].val.longval);
  char const *begin, *end;
  converter.convert(begin, end);
  
  if (begin == NULL) {
      setvarulong("error",1);
      setvarstring("perror", converter.error());
  }

  string ret(begin, end);
  back.returnval.stringval = ret;
  return back;
}
#include "Latin1ToUtf8.h"

int args_latin1toutf8[]= { TYPE_VALUE_STRING,
                           0 };

flowctrl func_latin1toutf8(type_block *aktblock,
                           opt *options,
                           joined_list<type_value> *args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back,call,call1;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror", "latin1toutf8: string not defined");
      return back;
    };

  Latin1ToUtf8 converter((*args)[0].stringval.c_str());
  char const *begin, *end;
  converter.convert(begin, end);
  string ret(begin, end);
  back.returnval.stringval = ret;
  return back;
}

int args_chr[]= { TYPE_VALUE_ULONG,
	              0 };

flowctrl func_chr(type_block* aktblock,opt* options,
			      joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  unsigned long const cp = (*args)[0].val.ulongval;
  if (cp > 0 && cp < 256) {
      char str[2];
      str[0] = (char)cp;
      str[1] = 0;
      back.returnval.stringval=str;
  } else {
      setvarulong("error",1);
      setvarstring("perror", "code point out of range 1..255");
  }
  return back;
}

