/*
BOIL (Brunni's Own Interpreter Language)
(c) 1998 netEstate (http://www.netestate.de/, info@netestate.de),
Michael Brunnbauer & Markus Hendel, Gesellschaft fuer Netzwerkkommunikation bR
Am Bluetenanger 8, 80995 Muenchen, Germany

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <tparser.h>

#include "general.h"
#ifdef WITH_RPC
#include "rpc.h"
#endif
#include "handles.cc"
#include "boil.h"
#include "boil.tab.h"
#include "libs.h"

int args_tcache_get[]= { TYPE_VALUE_STRING,
                         TYPE_VALUE_STRING,
                         TYPE_VALUE_STRING,
                         0 };

flowctrl func_tcache_get(type_block* aktblock,opt* options,
                         joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;
  
  if (!defined((*args)[0].stringval) ||
      !defined((*args)[1].stringval) ||
      !defined((*args)[2].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","tcache_get: string not defined");
      return back;
    };

  hashdir<joined_list<cachefield>,string,256,0>* table=
    tcache.get_listelem((*args)[0].stringval);
  if (!table) { return back; };

  joined_list<cachefield>* cols=table->get_listelem((*args)[1].stringval);
  if (!cols) { return back; };
  unsigned long colslen=cols->length();
  for (unsigned long i=0;i!=colslen;i++)
    {
      if ((*cols)[i].name==(*args)[2].stringval)
	{ back.returnval.stringval=(*cols)[i].value; break; };
    };
  return back;
}

int args_tcache_set[]= { TYPE_VALUE_STRING,
                         TYPE_VALUE_STRING,
                         TYPE_VALUE_STRING,
                         TYPE_VALUE_STRING,
                         0 };

flowctrl func_tcache_set(type_block* aktblock,opt* options,
                         joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_VOID;
  unsigned long i,u;

  if (!defined((*args)[0].stringval) ||
      !defined((*args)[1].stringval) ||
      !defined((*args)[2].stringval) ||
      !defined((*args)[3].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","tcache_set: string not defined");
      return back;
    };

  hashdir<joined_list<cachefield>,string,256,0>* table;

  /* cleaning up if using too much mem */
  if (string_mem>tcache_size)
    {
      /* increment actual index hash */
      tcache_hash++;
      
      /* hash-values for tables */
      for (i=0;i!=256;i++)
	{
	  /* tables for a hash value */
	  for (u=0;u!=tcache.values[i].length();u++)
	    {
              /* delete last index for actual hash */
	      table=&tcache.values[i][u];
	      unsigned long hashlen=table->names[tcache_hash].length();
	      if (hashlen)
		{
		  printf("deleting index %s in %s\n",
			      table->names[tcache_hash][hashlen-1].c_str(),
			      tcache.names[i][u].c_str());
		  fflush(stdout);
		  table->remove_listelem(table->names[tcache_hash][hashlen-1]);
		};
	    };
	};
    };
  
  table=tcache.get_listelem((*args)[0].stringval);
  if (!table)
    {
      table=new hashdir<joined_list<cachefield>,string,256,0>();
      if (!table) { return back; };
      tcache.insert_listelem(*table,(*args)[0].stringval);
      delete table;
      table=tcache.get_listelem((*args)[0].stringval);
    };

  joined_list<cachefield>* cols=
    table->get_listelem((*args)[1].stringval);
  if (!cols)
    { 
      cols=new joined_list<cachefield>();
      if (!cols) { return back; };
      table->insert_listelem(*cols,(*args)[1].stringval);
      delete cols;
      cols=table->get_listelem((*args)[1].stringval);
    };

  unsigned long colslen=cols->length();
  string* value=0;
  for (i=0;i!=colslen;i++)
    {
      if ((*cols)[i].name==(*args)[2].stringval)
	{ value=&(*cols)[i].value; break; };
    };

  if (value)
    {
      *value=(*args)[3].stringval;
    }
  else
    {
      cachefield cfield;
      cfield.name=(*args)[2].stringval;
      cfield.value=(*args)[3].stringval;
      cols->insert_listelem(cfield);
    };
  
  return back;
}

int args_tcache_del[]= { TYPE_VALUE_STRING,
                         TYPE_VALUE_STRING,
                         0 };

flowctrl func_tcache_del(type_block* aktblock,opt* options,
                         joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_VOID;

  if (!defined((*args)[0].stringval) ||
      !defined((*args)[1].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","tcache_del: string not defined");
      return back;
    };

  hashdir<joined_list<cachefield>,string,256,0>* table=
    tcache.get_listelem((*args)[0].stringval);
  if (!table) { return back; };

  table->remove_listelem((*args)[1].stringval);
  return back;
}

int args_tcache_dellike[]= { TYPE_VALUE_STRING,
			     TYPE_VALUE_STRING,
			     0 };

flowctrl func_tcache_dellike(type_block* aktblock,opt* options,
			     joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_VOID;

  if (!defined((*args)[0].stringval) ||
      !defined((*args)[1].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","tcache_dellike: string not defined");
      return back;
    };
  
  string search=(*args)[1].stringval;
  hashdir<joined_list<cachefield>,string,256,0>* table=
    tcache.get_listelem((*args)[0].stringval);
  if (!table) { return back; };

  for (unsigned long i=0;i!=256;i++)
    {
      for (unsigned long u=0;u!=table->names[i].length();)
	{
	  if (table->names[i][u].size()>=search.size() &&
	      !strncmp(table->names[i][u].c_str(),search.c_str(),search.size()))
	    {
	      table->remove_listelem(table->names[i][u]);
	      u=0;
	      continue;
	    };
	  u++;
	};
    };
  return back;
}

int args_tcache_setsize[]= { TYPE_VALUE_ULONG,
			     0 };

flowctrl func_tcache_setsize(type_block* aktblock,opt* options,
			     joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_VOID;

  tcache_size=(*args)[0].val.ulongval;
  return back;
}

int args_tcache_getsize[]= { 0 };

flowctrl func_tcache_getsize(type_block* aktblock,opt* options,
                             joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  
  back.returnval.val.ulongval=string_mem;
  return back;
}

string strtok_string1;

int args_htconvert[]= { TYPE_VALUE_STRING,
			0 };
flowctrl func_htconvert(type_block* aktblock,opt* options,
			joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","htconvert: string not defined");
      return back;
    };
  string c=(*args)[0].stringval;
  char buf1[2];
  string d,url,tag,tag1,token,zeile;
  unsigned long size,i;
  buf1[1]=0;

  /* alles ausser ASCII raus. warnungen einstreuen */
  d="";
  size=c.size();
  for (i=0;i!=size;i++)
    {
      unsigned char ch=c[i];
      buf1[0]=ch;
      if (ch==228) { d+="ae";}
      else if (ch==246) { d+="oe"; }
      else if (ch==252) { d+="ue"; }
      else if (ch==223) { d+="ss"; }
      else if (ch==196) { d+="Ae"; }
      else if (ch==214) { d+="Oe"; }
      else if (ch==220) { d+="Ue"; }
      else if (ch=='`') { d+="'"; }
      else if (ch=='\r') { ; }
      else if (ch>127) { d+=":SZ:"; }
      else {  d+=buf1; };
    };
  c=d;
  if (c=="") { back.returnval.stringval=c; return back; };

  /* returns und mehrfache returns/spaces entfernen */
  d="";
  strtok_string1=c;
  token=mystrtok(strtok_string1.lock()," \n");
  while (defined(token))
    {
      if (token!="") { d+=token+(string) " "; };
      token=mystrtok(0," \n");
    };
  c=d;
  if (c=="") { back.returnval.stringval=c; return back; };

  /* HTML-tags */
  d="";
  if (c[0]=='<') c=" "+c;
  strtok_string1=c;
  token=mystrtok(strtok_string1.lock(),"<");
  url="";
  while (defined(token))
    {
      d+=token;
      tag=mystrtok(0,">");
      if (!defined(tag)) { break; };
      tag1=tag;
      str_lwr(tag1.lock());
      token=mystrtok(0,"<");
      if (tag1=="p" ||
          strstr(tag1.c_str(),"ul")==tag1.c_str() ||
          strstr(tag1.c_str(),"/ul")==tag1.c_str() ||
          strstr(tag1.c_str(),"dir")==tag1.c_str() ||
          strstr(tag1.c_str(),"/dir")==tag1.c_str() ||
          strstr(tag1.c_str(),"ol")==tag1.c_str() ||
          strstr(tag1.c_str(),"/ol")==tag1.c_str() ||
          strstr(tag1.c_str(),"dl")==tag1.c_str() ||
          strstr(tag1.c_str(),"/dl")==tag1.c_str())
        {
          /* falls letztes zeichen space: loeschen */
          if (d.begin()[0] && d[d.size()-1]==' ')
            { d.lock()[d.size()-1]=0; d.unlock(); };

          /* falls ersten zeichen space: ueberspringen */
          if (defined(token) && token!="" && token[0]==' ')
            { token=token.c_str()+1; };

          d+="  ";
        }
      else if (strstr(tag1.c_str(),"li")==tag1.c_str() ||
               strstr(tag1.c_str(),"dt")==tag1.c_str() ||
               strstr(tag1.c_str(),"dd")==tag1.c_str())
        {
          /* falls letztes zeichen space: loeschen */
          if (d.c_str()[0] && d.c_str()[d.size()-1]==' ')
	    { d.lock()[d.size()-1]=0; d.unlock(); };

          /* falls ersten zeichen space: ueberspringen */
          if (defined(token) && token!="" && token[0]==' ')
            { token=token.c_str()+1; };

          d+="  -";
        }
      else if (strstr(tag1.c_str(),"a href=\""))
        {
          url=tag.c_str()+8;
          if (strstr(url.c_str(),"\""))
	    { strstr(url.lock(),"\"")[0]=0; url.unlock(); };
        }
      else if (tag1=="/a" && url!="")
        {
          /* falls letztes zeichen space: loeschen */
          if (d.c_str()[0] && d[d.size()-1]==' ') 
	    { d.lock()[d.size()-1]=0; d.unlock(); };

          d+=" ("+url+")";
          url="";
        };
    };
  c=d;
  if (c=="") { back.returnval.stringval=c; return back; };

  /* zeilenumbrueche neu machen */
  d="";
  strtok_string1=c;
  token=mystrtok(strtok_string1.lock()," \n");
  zeile="";
  while (defined(token))
    {
      if (token=="")
        {
          if (zeile=="") { d+="\n"; }
          else { d+=zeile+"\n\n"; zeile=""; };
        }
      else
        {
          if (zeile.size()+token.size() > 64) { d+=zeile+"\n"; zeile=""; };
          zeile=zeile+token+" ";
        };
      token=mystrtok(0," \n");
    };
  d+=zeile;
  c=d;
  if (c=="") { back.returnval.stringval=c; return back; };

  /* spaces und returns am ende wegmachen */
  i=c.size()-1;
  while (1)
    {
      char ch=c[i];
      if (ch!='\n' && ch!=' ') break;
      if (!i) break;
      i--;
    };
    
  c.lock()[i+1]=0;c.unlock();

  back.returnval.stringval=c;
  return back;
}

int args_findtag[]= { TYPE_VALUE_STRING,
		      TYPE_VALUE_ULONG,
		      TYPE_VALUE_ULONG,
		      0 };

flowctrl func_findtag(type_block* aktblock,opt* options,
		      joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0xffffffff;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","findtag: string not defined");
      return back;
    };

  char* c;
  char* w;
  char* t;
  unsigned long pos;
  type_value* var;
  flowctrl call;
  string search;

  search="code";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","findtag: cannot get variable code");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","findtag: variable code has wrong type");
      return back;
    };
  c=var->stringval.lock();

  search="werte";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","findtag: cannot get variable werte");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","findtag: variable werte has wrong type");
      return back;
    };
  w=var->stringval.lock();

  search="typen";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","findtag: cannot get variable typen");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","findtag: variable typen has wrong type");
      return back;
    };
  t=var->stringval.lock();

  pos=(*args)[2].val.ulongval;
  if (pos==0xffffffff)
    {
      setvarulong("error",1);
      setvarstring("perror","findtag: pos is 0xffffffff");
      return back;
    };

  if (!c[pos])
    {
      setvarulong("error",1);
      setvarstring("perror","findtag: EOF on ptr");
      return back;
    };

  tagcount=0;
  tagpos=0xffffffff;
  findtag((*args)[0].stringval,(*args)[1].val.ulongval,pos,c,w,t);
  back.returnval.val.ulongval=tagpos;
  return back;
};


int args_get_classelem[]= { TYPE_VALUE_STRING,
			    TYPE_VALUE_ULONG,
			    0 };

flowctrl func_get_classelem(type_block* aktblock,opt* options,
			    joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: string not defined");
      return back;
    };

  char* c;
  char* w;
  char* t;
  unsigned long pos;
  type_value* var;
  flowctrl call;
  string search;

  search="code";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable code");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable code has wrong type");
      return back;
    };
  c=var->stringval.lock();

  search="werte";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable werte");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable werte has wrong type");
      return back;
    };
  w=var->stringval.lock();

  search="typen";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable typen");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable typen has wrong type");
      return back;
    };
  t=var->stringval.lock();

  pos=(*args)[1].val.ulongval;

  string err;
  back.returnval.val.ulongval=classelem(c,w,t,(*args)[0].stringval.lock(),pos,err);
  if (defined(err))
    {
      setvarulong("error",1);
      setvarstring("perror",err);
    };
  return back;
}

int args_get_ulongelem[]= { TYPE_VALUE_STRING,
                            TYPE_VALUE_ULONG,
                            0 };

flowctrl func_get_ulongelem(type_block* aktblock,opt* options,
                            joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: string not defined");
      return back;
    };

  char* c;
  char* w;
  char* t;
  unsigned long pos;
  type_value* var;
  flowctrl call;
  string search;

  search="code";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable code");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable code has wrong type");
      return back;
    };
  c=var->stringval.lock();

  search="werte";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable werte");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable werte has wrong type");
      return back;
    };
  w=var->stringval.lock();

  search="typen";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable typen");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable typen has wrong type");
      return back;
    };
  t=var->stringval.lock();

  pos=(*args)[1].val.ulongval;

  string err;
  back.returnval.val.ulongval=ulongelem(c,w,t,(*args)[0].stringval.lock(),pos,err);
  if (defined(err))
    {
      setvarulong("error",1);
      setvarstring("perror",err);
    };
  return back;
}

int args_get_longelem[]= { TYPE_VALUE_STRING,
			   TYPE_VALUE_ULONG,
			   0 };

flowctrl func_get_longelem(type_block* aktblock,opt* options,
			   joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_LONG;
  back.returnval.val.ulongval=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: string not defined");
      return back;
    };

  char* c;
  char* w;
  char* t;
  unsigned long pos;
  type_value* var;
  flowctrl call;
  string search;

  search="code";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable code");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable code has wrong type");
      return back;
    };
  c=var->stringval.lock();

  search="werte";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable werte");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable werte has wrong type");
      return back;
    };
  w=var->stringval.lock();

  search="typen";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable typen");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable typen has wrong type");
      return back;
    };
  t=var->stringval.lock();

  pos=(*args)[1].val.ulongval;

  string err;
  back.returnval.val.longval=longelem(c,w,t,(*args)[0].stringval.lock(),pos,err);
  if (defined(err))
    {
      setvarulong("error",1);
      setvarstring("perror",err);
    };
  return back;
}

int args_get_stringelem[]= { TYPE_VALUE_STRING,
			     TYPE_VALUE_ULONG,
			     0 };

flowctrl func_get_stringelem(type_block* aktblock,opt* options,
			     joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: string not defined");
      return back;
    };

  char* c;
  char* w;
  char* t;
  unsigned long pos;
  type_value* var;
  flowctrl call;
  string search;

  search="code";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable code");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable code has wrong type");
      return back;
    };
  c=var->stringval.lock();

  search="werte";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable werte");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable werte has wrong type");
      return back;
    };
  w=var->stringval.lock();

  search="typen";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable typen");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable typen has wrong type");
      return back;
    };
  t=var->stringval.lock();

  pos=(*args)[1].val.ulongval;

  string err;
  back.returnval.stringval=stringelem(c,w,t,(*args)[0].stringval.lock(),pos,err);
  if (defined(err))
    {
      setvarulong("error",1);
      setvarstring("perror",err);
    };
  return back;
}

int args_get_ushortelem[]= { TYPE_VALUE_STRING,
			     TYPE_VALUE_ULONG,
			     0 };

flowctrl func_get_ushortelem(type_block* aktblock,opt* options,
			     joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.ulongval=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: string not defined");
      return back;
    };

  char* c;
  char* w;
  char* t;
  unsigned long pos;
  type_value* var;
  flowctrl call;
  string search;

  search="code";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable code");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable code has wrong type");
      return back;
    };
  c=var->stringval.lock();

  search="werte";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable werte");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable werte has wrong type");
      return back;
    };
  w=var->stringval.lock();

  search="typen";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable typen");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable typen has wrong type");
      return back;
    };
  t=var->stringval.lock();

  pos=(*args)[1].val.ulongval;

  string err;
  back.returnval.val.ulongval=ushortelem(c,w,t,(*args)[0].stringval.lock(),pos,err);
  if (defined(err))
    {
      setvarulong("error",1);
      setvarstring("perror",err);
    };
  return back;
}

int args_get_shortelem[]= { TYPE_VALUE_STRING,
                            TYPE_VALUE_ULONG,
			    0 };

flowctrl func_get_shortelem(type_block* aktblock,opt* options,
			    joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_LONG;
  back.returnval.val.ulongval=0;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: string not defined");
      return back;
    };

  char* c;
  char* w;
  char* t;
  unsigned long pos;
  type_value* var;
  flowctrl call;
  string search;

  search="code";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable code");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable code has wrong type");
      return back;
    };
  c=var->stringval.lock();

  search="werte";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable werte");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable werte has wrong type");
      return back;
    };
  w=var->stringval.lock();

  search="typen";
  call=getvar(aktblock,options,var,search);
  if (call.ctrl!=FLOW_OK)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: cannot get variable typen");
      return back;
    };
  if (var->type!=TYPE_VALUE_STRING)
    {
      setvarulong("error",1);
      setvarstring("perror","get_*: variable typen has wrong type");
      return back;
    };
  t=var->stringval.lock();

  pos=(*args)[1].val.ulongval;

  string err;
  back.returnval.val.longval=shortelem(c,w,t,(*args)[0].stringval.lock(),pos,err);
  if (defined(err))
    {
      setvarulong("error",1);
      setvarstring("perror",err);
    };
  return back;
}

int args_templateparser[]= { TYPE_VALUE_STRING,
                             0 };

static type_value &TemplateParserLookupVar(char const *varName,
                                           bool &found,
                                           void* aktblock,
                                           void* options)
{
    type_value *tokens;
    string strVarName = varName;
    flowctrl call = getvar(static_cast<type_block*>(aktblock),
                           static_cast<opt*>(options),
                           tokens, strVarName);

    if (call.ctrl != FLOW_OK)
    {
        found = false;
        static type_value emptyStr;
        emptyStr.type = TYPE_VALUE_STRING;
        emptyStr.stringval = "";
        return emptyStr;
    }
    found = true;
    return *tokens;
}

static flowctrl TemplateParserExecFunc(string funcName,
                                       void* aktblock,
                                       void* options,
                                       joined_list<type_value> &args) {
    type_function *function;
    flowctrl call = getfunc(static_cast<type_block*>(aktblock),
                            static_cast<opt*>(options),
                            function, funcName);
    
    if (call.ctrl == FLOW_OK)
      {
	if (function->block==(type_block*) 1)
	  {
	    /* bad hack for system function recursion */
	    type_block* ptr=static_cast<type_block*>(aktblock);
	    bool alert=0;
	    while (ptr)
	      {
		if (ptr==function->myblock) 
		  {
		    alert=1;
		    function->block=ptr->block;
		    break;
		  };
		ptr=ptr->block;
	      };
	    
	    /* upper block of (normal) system function is calling block */
	    if (!alert) function->block=static_cast<type_block*>(aktblock);
	  };
	
	return function->call(static_cast<type_block*>(aktblock),
			      static_cast<opt*>(options), false, false,
			      &args);
      }
    return call;    
}

flowctrl func_templateparser(type_block* aktblock,opt* options,
                             joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_STRING;

  if (!defined((*args)[0].stringval))
    {
      setvarulong("error",1);
      setvarstring("perror","templateparser: template string not defined");
      return back;
    };

  TemplateParser parser(TemplateParserLookupVar,
                        TemplateParserExecFunc,
                        aktblock, options);

  parser.SetSrc((*args)[0].stringval);
  parser.Parse();
  bool error_occured;
  bool flow_error_occured;
  std::string error_msg;
  back.returnval.stringval = parser.GetResult(error_occured, flow_error_occured, error_msg);

  if (error_occured) {
      ::setvarulong("error", 1);
      ::setvarstring("perror", error_msg.c_str());
      if (flow_error_occured) {
          return parser.GetFlowError();
      }
  }

  return back;
}

int args_delayed_query[]= { TYPE_VALUE_STRING,
                            TYPE_VALUE_STRING,
                            0 };

flowctrl func_delayed_query(type_block* aktblock,opt* options,
                            joined_list<type_value>* args)
{
  aktblock=aktblock;options=options;args=args; /* to prevent warnings */
  flowctrl back;
  back.ctrl=FLOW_RETURN;
  back.returnval.type=TYPE_VALUE_ULONG;
  back.returnval.val.longval = 0;

  string &fifo_path = (*args)[0].stringval;

  if (!defined(fifo_path)) {
      setvarulong("error", 1);
      setvarstring("perror", "delayed_query: fifo_path not defined");
      return back;
  }

  string &query = (*args)[1].stringval;

  if (!defined(query)) {
      setvarulong("error", 1);
      setvarstring("perror", "delayed_query: query not defined");
      return back;
  }

  int const msg_size = query.size() + 16;

  if (msg_size + 1 > PIPE_BUF) {
      setvarulong("error", 1);
      setvarstring("perror", "delayed_query: query too long");
      return back;
  }

  char msg[PIPE_BUF];
  sprintf(msg, "\xFF\xFF\xFF\xFF\xFF%06d\xFE\xFE\xFE\xFE\xFE%s",
               query.size(), query.c_str());

  int fd = open(fifo_path.c_str(), O_WRONLY);
  if (fd == -1) {
      setvarulong("error", 1);
      setvarstring("perror", "delayed_query: error opening fifo");
      return back;
  }
  write(fd, msg, msg_size);
  close(fd);
  back.returnval.val.longval = 1;
  return back;
}
